/***************************************************************************
                          ImageBox.h  -  description
                             -------------------
    begin                : ven mag 7 2004
    copyright            : (C) 2004 by Daniele Lavatelli
    email                : ldanilinux@interfree.it
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef IMAGE_BOX_H

 #define IMAGE_BOX_H

 #include <qimage.h>
 #include <qpopupmenu.h>
 #include <qevent.h>
 #include <qpixmap.h>
 #include <qlabel.h>
 #include <qframe.h>
 #include <pthread.h>

 #include "Net.h"

 //Unselected - no errors background color
 extern QColor NOTSEL_NOERRORS;

 //Unselected - warning background color
 extern QColor NOTSEL_WARNING;

 //Unselected - critical background color
 extern QColor NOTSEL_CRITICAL;

 //Selected - no errors background color
 extern QColor SEL_NOERRORS;

 //Selected - warning background color
 extern QColor SEL_WARNING;

 //Selected - critical background color
 extern QColor SEL_CRITICAL;

 ///Defines the error state of the ImageBox class
 enum ErrorType  {ERR_WARNING,ERR_CRITICAL,ERR_NOERRORS};
 
 ///This class handles connection and frame rendering
 class ImageBox : public QWidget
 {

  Q_OBJECT
  
  public:
  ///creates the ImageBox
  /**\param[in] x x coordinate of this ImageBox in the MainWindow's grid
  \param[in] y y coordinate of this ImageBox in the MainWindow's grid*/
  ImageBox(QWidget *,const char *,int x,int y);
  ///Destroys the ImageBox
  ~ImageBox();
  ///Selects this ImageBox
  /**When ImageBox is selected, it emits a StatusMsg() signal its state changes*/
  void Select();
  ///Deselects this ImageBox
  void Deselect();
  ///Autoconnects the server
  /**\param[in] name The server host name
  \param[in] port The server port
  
  This function is a shortcut: it connects the default server whithout calling the Connect() slot (so, without siplaying the "Connect to server" dialog.*/
  void AutoConnect(const char *name,const char *port);
  ///Used to resize the ImageBox to fit the received frame
  /**Only used in the autoconnect mode.*/
  void AdjustSize();
  ///Returns the name of the server which is currently handled by this ImageBox
  /**\return A string containing the server name*/
  char *GetServerName();
  ///Strats the drawing thread
  /**\param data Pointer to this ImageBox. Internal use only.*/
  static void* StartThread(void* data);

  signals:
   ///Emitted when the error state changes
   /**\param[out] err The error status
   \param[out] msg The status message
   
   This message is emitted in selected staus only*/
   void StatusMsg(ErrorType err,char *msg);
   ///Emitted when the user clicks to select this ImageBox
   /**\param x x coordinate of this ImageBox on the MainWindow's grid
   \param y y coordinate of this ImageBox on the MainWindow's grid*/
   void Clicked(int x,int y);
   ///Emitted about once a second. Lets MainWindow display the current frame rate.
   /**\param[out] fps The frame rate*/
   void FrameRateMsg(float fps);
  
  private:
  QPopupMenu      *mpMenu;///<Defines the main menu. /**This menu handles server connection/disconnection and frame resizing method*/
  QImage          mImage;///<The main frame
  QImage          temp_frame;///<The auxiliary frame
  QWidget         *mpBox;///<The widget which contains the frame
  JpegFrameType   mJpegFrame;///<It stores the jpeg frame
  Connection      mConnection;///<It handles the connection
  ///The main rendering loop
  /**This function continuously downloads frames from the server and posts a TimerEvent which renders them from the main application thread.
  Simultaneous reads and writes on the frame data (stored in mImage) are avoided by using the mBusy mutex.*/
  void            Loop();
  ///Sets the error status of this ImageBox and updates its current status message
  /**\param err The new error status
  \param msg The new status message*/
  void            SetErrorStatus(ErrorType err,char *msg);
  pthread_t       mThId;///<Rendering thread structure
  pthread_mutex_t mBusy;///<Rendering thread mutex
  bool            mDrawActive;///<This flag controls the drawing loop
  int             stretch_id;///<index of the "stretch" item in the main menu
  int             fit_id;///<index of the "fit-keep ratio" item in the main menu
  int             normal_id;///<index of the "original size" item in the main menu
  int             connect_id;///<index of the "connect" item in the main menu
  int             reconnect_id;///<index of the "reconnect" item in the main menu
  int             disconnect_id;///<index of the "disconnect" item in the main menu
  int             inspect_id;///<index of the "inspect" item in the main menu
  ErrorType       mErrorStatus;///<The error status of ImageBox
  char            mStatusMessage[255];///<The message displayed in the MainWindow's StatusBar when this ImageBox is selected
  bool            mSelected;///<\em true if this is the selected ImageBox, \em false otherwise
  int             mX;///<x coordinate of this ImageBox in the MainWindow's grid
  int             mY;///<y coordinate of this ImageBox in the MainWindow's grid
  char            mServerName[255];///<Name of the server which is currently handled by this ImageBox
  char            mServerPort[10];///<Port of the server which is currently handled by this ImageBox
  
  private slots:
  ///Switches to the "stretch" draw mode
  void ShowStretch();
  ///Switches to the "fit-keep ratio" draw mode
  void ShowFit();
  ///Switches to the "original size" draw mode
  void ShowNormal();
  ///Shows the "Connect to server" dialog and connects a new server
  void Connect();
  ///Reconnects the last connected server
  void Reconnect();
  ///Disconnects the server
  /**ImageBox saves the last server's address, to let the user reconnect it if he wants.*/
  void Disconnect();
  ///-EXPERIMENTAL- not implemented
  void Inspect();

  protected:
  ///Handles mouse clicks on this ImageBox
  /**A right click selects this ImageBox and shows the main menu.
  
  A left click just selects this ImageBox.
  \param[in] e The mouse event - automatically sent by Qt*/
  void mousePressEvent( QMouseEvent *e );
  ///Handles the resize event when the size of this ImageBox changes.
  /**\param[in] e The resize event automatically sent by Qt*/
  void resizeEvent( QResizeEvent *e );
  ///Redraws the frame
  /**\param e The event
  
  This event is always generated by ImageBox when it needs to redraw its area. \n There are 3 events: \n
  EVENT_REDRAWFRAME redraws the last frame stored in temp_frame. Used when the current frame is unavailable (i.e. when the server has been disconnected) \n
  EVENT_DRAWFRAME is the standard event. It waits frame to be downloaded from the net, then resizes and draws it \n
  EVENT_REDRAWBACKGROUND redraws the background. It is used when ImageBox's error status or selection status (=> background color) changes.*/
  void timerEvent( QTimerEvent *e );
  ///Handles redrawing of the ImageBox
  /**\param[in] e The redraw event automatically sent by Qt*/
  void paintEvent( QPaintEvent *e );
  
 };

 
#endif
