/*
    Network-related stuff.
 
    Copyright 2004 by David Santinoli <david@santinoli.com>
 
    Translated to C++ by Daniele Lavatelli <ldaniele@interfree.it>
*/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "Net.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

static const char rcsid[]="$Id: Net.cpp,v 1.2 2004/07/03 13:20:57 santinod Exp $";

void Connection::Log(char *err, int type=ERR)
{
 char stre[256];

 if(!(type&ERR) || (type&NOSTRERR)) {
    snprintf(mLastError,ERRBUFLEN,"%s",err);
    }
 else {
    /* strerror_r() is currently broken, let's stick to strerror() */
    strncpy(stre,strerror(errno),256-1);
    stre[255]=0;
    snprintf(mLastError,ERRBUFLEN,"%s (%s)",err,stre);
    }
 fprintf(stderr,"%s\n",mLastError);
}

void Connection::ResetErr(void)
{
 mLastError[0]=0;
}

Connection::Connection()
{
 mConnStatus = CONN_UNCONNECTED;
 greeting=0;
 snapshot=0;
 ResetErr();
}

Connection::~Connection()
{
 if (mConnStatus != CONN_UNCONNECTED) Disconnect();
}

/*
   Read up to \r\n into *buf.
   Return 0 upon disconnection, -1 upon read() failure,
   -2 upon buffer overflow, >0 upon reading a well-formed line.
*/
int Connection::net_read_line(char *buf, int len)
{
 int i=0, n;

 while(i<len) {
    if (mConnStatus != CONN_CURRENT) return 0; //disconnected
    if( (n=read(mSock,buf+i,1))<=0) {
       if(n<0) Log("Read error (headers)");
       return (n<0) ? -1 : 0;
       }
    if(i>0 && buf[i]=='\n' && buf[i-1]=='\r') {
       buf[i-1]=0;
       return i+1;
       }
    i++;
    }
 Log("Buffer overflow while reading headers!",NOSTRERR);
 return -2;
}

int Connection::HttpRequest(const char * user_agent)
{
 char http_request[100];
 int n;

 strcpy(http_request,"GET /\r\nUser-Agent: ");
 strncat(http_request,user_agent,50);
 strcat(http_request,"\r\n\r\n");
 n=strlen(http_request);
 if(n!=write(mSock,http_request,strlen(http_request))) {
    Log("error writing HTTP headers");
    return -1;
    }

 
 return 0;
}

int Connection::Disconnect(void)
{
 if(mConnStatus==CONN_UNCONNECTED) return 0;

 mConnStatus=CONN_UNCONNECTED;
 if(close(mSock)) {
    Log("close()");
    return -1;
    }
 ResetErr();
 if(greeting) { free(greeting); greeting=0; }
 snapshot=0;
 return 0;
}

int Connection::Connect(const char *server, const char *port)
{
 struct addrinfo *ai, hai;
 int n;

 if(mConnStatus!=CONN_UNCONNECTED) return 1;

 hai.ai_flags=0;
 hai.ai_family=PF_INET;
 hai.ai_socktype=SOCK_STREAM;
 hai.ai_protocol=0;

 
 if((n=getaddrinfo(server,port,&hai,&ai))) {
    Log("getaddrinfo()");
    return -1;
    }

 mSock=socket(PF_INET,SOCK_STREAM,0);
 if(mSock<0) {
    Log("socket()");
    return -1;
    }

 mConnStatus=CONN_CONNECTING;

 if(connect(mSock,ai->ai_addr,ai->ai_addrlen)) {
    Log("connect()");
    return -1;
    }

 mConnStatus=CONN_CURRENT;

 ResetErr();
 return 0;
  
}

int Connection::ParseGlobalHeaders(void)
{
 while(net_read_line(mBuf,BUFLEN)>2) {
    if(!strncmp("X-Greeting: ",mBuf,12)) {
       greeting=strdup(mBuf+12);
       fprintf(stderr,"Server greeting: %s\n",greeting);
       }
    else if(!strcmp("X-Snapshot: 1",mBuf)) {
       snapshot=1;
       break;
       }
    }
 ResetErr();
 return 0;
}

int Connection::Ack(char *s, int len)
{
 if (mConnStatus == CONN_UNCONNECTED) return 0; 
 if(write(mSock,s,len)<=0) {
    Log("write (ack)");
    return -1;
    }

 ResetErr();
 return 0;
}

/*
   Read local headers and image data.
   Return 0 upon disconnection, <0 upon error, 1 upon success.
*/
int Connection::GetImage(JpegFrameType *jframe)
{
 int n;
 unsigned long t;
 char *c;

 if(!snapshot) do {
    if( (n=net_read_line(mBuf,BUFLEN))<=0 ) return n;
    } while(strcmp(mBuf,MP_BOUNDARY));

 /* Content-Type: */
 do {
    if( (n=net_read_line(mBuf,BUFLEN))<=0 ) return n;
    if(!(c=strchr(mBuf,':'))) break;
    *c=0;
    if(!strcmp(mBuf,"Content-Length")) jframe->length=atoi(c+2);
    else if(!strcmp(mBuf,"X-Index")) jframe->index=atoi(c+2);
    } while(n>2);

 if(n!=2) {
    Log("Malformed frame headers",NOSTRERR);
    return -1;
    }

 /* Buffer allocation */
 while(jframe->length>jframe->jpeg_buffer_length) {
    if( (jframe->jpeg_buffer=(unsigned char*)realloc(jframe->jpeg_buffer,
       jframe->jpeg_buffer_length*=2)) ) {
       Log("realloc()ed frame buffer",LOG);
       }
    else {
       Log("cannot realloc() frame buffer",NOSTRERR);
       return -1;
       }
    }  

 t=0;
 while(t<jframe->length) {
    if (mConnStatus == CONN_CURRENT) n=read(mSock,jframe->jpeg_buffer+t,jframe->length-t); else return 0;
    if(n<=0) {
       fprintf(stderr,"read(): %d\n",n);
       if(n<0) Log("Read error (frame)");
       return n;
       }
    t+=n;
    }

 ResetErr();
 return 1;
}
