/*
   The Palantir server

   Copyright 2000-2007 by David Santinoli <david@santinoli.com>

   IPv6 support initially contributed by
   Petri Nyknen <petri.t.nykanen@kolumbus.fi>,
   inspired by "Porting applications to IPv6 HowTo" by Eva M. Castro at
   http://jungla.dit.upm.es/~ecastro/IPv6-web/ipv6.html

*/

#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <netdb.h>
#include <linux/videodev.h>
#include "config.h"
#include "diag.h"
#include "video.h"
#include "audio.h"
#include "net.h"
#include "jpeg.h"
#include "telmu.h"
#include "main.h"
#include "queue.h"
#include "parser.h"
#include "sensor_def.h"
#include "caption.h"
#include "x11mon.h"

/* The following is missing from MIPS' includes */
#ifndef IPV6_V6ONLY
#define IPV6_V6ONLY              26
#endif

#define VIDEO_UNAVAIL_JPG "unavail.jpg"

#define DEBUG_DUMP_IMG_GLOBAL	0
#define DEBUG_DUMP_IMG_LOCAL	0
#define DEBUG_DUMP_CLIENT_DATA	0
#define DEBUG_ALLOW_ALL_AUDIO_REQS	0

#define AUDIO_ENABLED (config.noaudio==0 && (config.audio_in_device || config.audio_out_device))

static const char rcsid[]="$Id: main.c,v 1.33 2007/05/18 23:19:14 santinod Exp $";

pthread_mutex_t jpeg_buf_mtx=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t jpeg_buf_size_mtx=PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t new_frame_cond=PTHREAD_COND_INITIALIZER;
int viewer_only=0,nojpeg=0,novideo=0;
int all_off=0;

static pthread_mutex_t clients_mtx=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t kill_capture_mtx=PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t kill_capture_cond=PTHREAD_COND_INITIALIZER;
static int clients=0;
int current_frame=0;

struct namevalue { char *name; char *value; };


int memcpy_realloc(void *src, void *dest, unsigned long int srcsize, unsigned long int *destsize)
{
 void *p;
 void **d=(void **)dest;

 while(srcsize>*destsize) {
    if( (p=realloc(*d,*destsize*2))==NULL ) {
       diag(SS_MAIN|ERROR,"Cannot realloc() JPEG buffer\n");
       return 1;
       }
    *d=p;
    *destsize*=2;
    diag(SS_MAIN|DEBUG,"Local JPEG buffer resized to %lu (need %lu)\n",*destsize,srcsize);
    }
 memcpy(*d,src,srcsize);
 return 0;
}


/*
   read at most len chars from fd until \r\n is found. Return string length
   (>=2), -1 on read error (or buffer overflow), -2 on select() timeout
*/
static int read_line(int fd, char *buf, int len, int timeout)
{
 int i=0, n, s;
 fd_set set;
 struct timeval to;

 FD_ZERO(&set);
 FD_SET(fd,&set);
 to.tv_sec=timeout;
 to.tv_usec=0;

 while(i<len) {
    do {
       s=select(fd+1,&set,0,0,&to);
       if(s==0) return -2;
       if(s<0 && errno!=EINTR) {
          diag(SS_MAIN|ERR,"select() in read_line failed");
          return -1;
          }
       } while(s<0);
    if( (n=read(fd,buf+i,1))<=0) break;
    if(i>0 && buf[i]=='\n' && buf[i-1]=='\r') return i+1;
    i++;
    }
 return (i==len)?0:-1;
}


static int create_sock(unsigned short int port, int family)
{
 int x=1;
 struct addrinfo hints, *res, *ressave;
 int n, sockfd;
 char service[NI_MAXSERV];
 char srv_addr[NI_MAXHOST];

 memset(&hints,0,sizeof(struct addrinfo));
 snprintf(service,sizeof(service),"%d",port);

 /* See explanation for AI_PASSIVE in 
    http://jungla.dit.upm.es/~ecastro/IPv6-web/ipv6.html
  */

 hints.ai_flags=AI_PASSIVE;

 /* a bit paranoid approach, but ensure sane value */
 hints.ai_family=AF_UNSPEC;
 if(family==AF_INET) hints.ai_family=AF_INET;
 if(family==AF_INET6) hints.ai_family=AF_INET6;

 hints.ai_socktype=SOCK_STREAM;

 n=getaddrinfo(NULL,service,&hints,&res);
 if(n<0) diag(SS_MAIN|DIERR,"getaddrinfo");
 
 ressave=res;

 /* Try open socket with each address until getting a valid listening socket
  */
 sockfd=-1;
 while(res) {
    sockfd=socket(res->ai_family,res->ai_socktype,res->ai_protocol);
    if(!(sockfd<0)) {
        if(setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,&x,sizeof(x)))
           diag(SS_MAIN|DIERR,"setsockopt() (SO_REUSEADDR)");
        if(family==AF_INET6 && res->ai_family==AF_INET6 && setsockopt(sockfd,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x)))
           diag(SS_MAIN|DIERR,"setsockopt() (IPV6_V6ONLY)");
        if(bind(sockfd,res->ai_addr,res->ai_addrlen)==0)
           break;
        close(sockfd);
        sockfd=-1;
        }
    res=res->ai_next;
    }

 if(sockfd<0) {
    freeaddrinfo(ressave);
    diag(SS_MAIN|DIERR,"socket() - could not open socket");
    return -1;
    }

 if(getnameinfo(res->ai_addr,res->ai_addrlen,srv_addr,sizeof(srv_addr),0,0,
    NI_NUMERICHOST|NI_NUMERICSERV)!=0) {
    strcpy(srv_addr,"unknown");
    }

 diag(SS_MAIN|INFO,"Listening on %s port %d\n",srv_addr,port);

 freeaddrinfo(ressave);
 return sockfd;
}


static void release_control()
{
 if(aud_status==A_CURRENT) audio_terminate_threads();
 aud_status=A_MUTE;
}


static struct namevalue parse_client_response(char **s)
{
 char *name=*s, *next, *value;
 struct namevalue nv={0, 0};
 
 while(*name==' ') name++;
 if(!*name) return nv;
 next=strchr(name,';');
 if(next) *next=0;
 nv.name=name;
 value=strchr(name,'=');
 if(value) {
    *value++=0;
    nv.value=value;
    }
 /* when '=' or value is missing, invalidate returned structure */
 if(!value || !*value) nv.name=0;
 if(next) *s=++next;
 else *s=0;
 return nv;
}


void decrement_clients(void)
{
 pthread_mutex_lock(&clients_mtx);
 --clients;
 if(!novideo)
    if(!clients) {
       pthread_mutex_lock(&kill_capture_mtx);
       pthread_cond_wait(&kill_capture_cond,&kill_capture_mtx);
       pthread_mutex_unlock(&kill_capture_mtx);
       }
 pthread_mutex_unlock(&clients_mtx);
 diag(SS_MAIN|DEBUG,"%d client(s) connected\n",clients);
}


void dump_sensors(int fd)
{
 int *sensor_current, i;
 FILE *f;
 time_t now;
 char timestamp[60];
 char *head="<HTML>\n<HEAD>\n<TITLE>Device status</TITLE>\n"
            "<BODY BGCOLOR=\"white\">\n";
 char *tail="</BODY>\n</HTML>\n";
 float scale=1, offset, v;

 if(!(f=fdopen(fd,"w"))) diag(SS_HTTP|DIE,"fdopen()");

 fprintf(f,"%s",head);
 fprintf(f,"<CENTER><H2>%s</H2></CENTER>\n",config.greeting);

 time(&now);
 if(strftime(timestamp,60,"%A %e %B %Y &nbsp;&nbsp;&nbsp; %T",localtime(&now)))
 fprintf(f,"<CENTER><H3>%s</H3></CENTER>\n",timestamp);

 if(!sensors) {
    fprintf(f,"<CENTER><H2>No sensors defined!</H2></CENTER>\n");
    }
 else {

 sensor_current=malloc(sensors*sizeof(int));

 pthread_mutex_lock(&telmu_io2http_mtx);
 for(i=0;i<sensors;i++)
    sensor_current[i]=sensor[i].value;
 pthread_mutex_unlock(&telmu_io2http_mtx);

 fprintf(f,"<TABLE WIDTH=100%% CELLSPACING=2 CELLPADDING=3>\n");
 for(i=0;i<sensors;i++) {
    if(sensor[i].imax-sensor[i].imin)
       scale=(atof(sensor[i].cfmax)-atof(sensor[i].cfmin))/
          (sensor[i].imax-sensor[i].imin);
    offset=atof(sensor[i].cfmin)-scale*sensor[i].imin;
    v=scale*sensor_current[i]+offset;
    fprintf(f,"<TR BGCOLOR=%s>\n<TD><B>%s</B><BR><I>%s</I>\n<TD ALIGN=RIGHT>%g\n",
       (i&1?"#f8f8f8":"#e0e0e0"),
       *sensor[i].name?sensor[i].name:"&nbsp;",
       *sensor[i].hint?sensor[i].hint:"&nbsp;",v);
    }
 fprintf(f,"</TABLE>\n");

 free(sensor_current);
    }

 fprintf(f,"<BR><BR>\n<CENTER><SMALL>Powered by <A HREF=\"http://www.fastpath.it/products/palantir/\">"VERSION"</A></SMALL></CENTER>\n");
 fprintf(f,"%s",tail);
 fclose(f);
}


void *http_server(struct conn_parms *cops)
{
 unsigned long int sent_frames=0;
 int stream=1, read_frame=-1, dup_frames=0;
 int is_dup=0, need_ack=0, i, h_lines=0, blanking=0;
 int sensor_only=0;
 unsigned long int my_jpeg_bufsize;
 unsigned long int jbytes=0;
 unsigned long long int tbytes=0;
 float dtime;
 enum control_status ctl_status;
 struct timeval tv1,tv2;
 char *boundary="\r\n--mp-boundary\r\n";
 char *nshead="HTTP/1.0 200 OK\r\nContent-Type: multipart/x-mixed-replace; boundary=--mp-boundary\r\n";
 char *iehead="HTTP/1.0 200 OK\r\n";
 char *ctype="Content-Type: image/jpeg\r\n";
 char *nocache="Expires: Thu, 19 Nov 1981 08:52:00 GMT\r\n"
    "Pragma: no-cache\r\n"
    "Cache-Control: no-store, no-cache\r\n";
 char *xlen="Content-Length: %d\r\n";
 char *xidx="X-Index: %d\r\n";
 char *proto_version="X-Protocol-Version: " PROTOCOL_VERSION "\r\n";
 char rbuf[1024];
 unsigned char *my_jpeg_buf;
 char headbuf[256];
 char *c, ctl_hdr[160];
 char peer_addr[NI_MAXHOST];
 pthread_t self;
 time_t ctl_expiry=0;
 int *sensor_current, *sensor_fresh, local_seqno=0, sensor_update=0;
 char sensor_hdr[250], sensor_tmp[20];
 char decoded_text[128];

 pthread_mutex_lock(&jpeg_buf_size_mtx);
 my_jpeg_bufsize=jpeg_buf_size;
 pthread_mutex_unlock(&jpeg_buf_size_mtx);

 my_jpeg_buf=malloc(my_jpeg_bufsize);
 sensor_fresh=malloc(sensors*sizeof(int));
 sensor_current=malloc(sensors*sizeof(int));
 self=pthread_self();
 ctl_status=C_DEQUEUED;

 if(getnameinfo((struct sockaddr *)&(cops->sa),sizeof(cops->sa),
    peer_addr,sizeof(peer_addr),0,0,
    NI_NUMERICHOST|NI_NUMERICSERV)!=0) {
    strcpy(peer_addr,"unknown");
    }

 diag(SS_HTTP|DEBUG,"%s - video server started\n",peer_addr);

 /* we sample time here because the second gettimeofday(&tv1,...) later on
    might not be reached, giving troubles with tv2-tv1 */
 if(gettimeofday(&tv1,NULL)) diag(SS_HTTP|DIERR,"gettimeofday");

 /* parse HTTP headers, looking for interesting stuff */
 do {
    i=read_line(cops->fd,rbuf,1024,15);
    if(i<=0) {
       if(i==-2) diag(SS_HTTP|DEBUG,"%s - timeout waiting for HTTP headers\n",peer_addr);
       goto out;
       }

    rbuf[i-2]=0;

    if(h_lines++==0) {
       /*diag(SS_HTTP|DEBUG,"%s: %s\n",peer_addr,rbuf);*/
       if(!strncmp(rbuf,"GET /",5)) {
          char *sp;
          if((sp=strchr(rbuf+5,' '))) *sp=0;
          diag(SS_HTTP|DEBUG,"%s - requested URL `%s'\n",peer_addr,rbuf+4);
          if(!strcmp(rbuf+5,"?mode=single")) stream=0;
          else if(!strcmp(rbuf+5,"?mode=devices")) sensor_only=1;
          else if(rbuf[5]) {
             diag(SS_HTTP|DEBUG,"%s - illegal URL request\n",peer_addr);
             write(cops->fd,"Illegal URL request\r\n",21);
             goto out;
             }
          }
       }
    if(!strncmp(rbuf,"User-Agent: ",12)) {
       char *ua=rbuf+12; /* strlen("User-Agent") */
       diag(SS_HTTP|INFO,"%s - %s\n",peer_addr,rbuf);
       if(strstr(ua,"pclient/ack")) need_ack=1;

       /* MSIE/Mac supports multipart/replace */
       if( (c=strstr(ua,"MSIE")) && !strstr(c,"Mac") ) stream=0;

       /* Opera doesn't. */
       if(strstr(ua,"Opera")) stream=0;
       }
    } while (i>2);

 if(sensor_only) { dump_sensors(cops->fd); goto out; }
 if(novideo && !need_ack) stream=0;

 if(cops->clients>config.max_clients) stream=0;

 diag(SS_HTTP|INFO,"%s - streaming mode: %s, %s loop\n",peer_addr,(stream?"continuous":"single-shot"),(need_ack?"closed":"open"));

 if(stream) write(cops->fd,nshead,strlen(nshead));
 else write(cops->fd,iehead,strlen(iehead));
 write(cops->fd,nocache,strlen(nocache));

 write(cops->fd,proto_version,strlen(proto_version));
 snprintf(headbuf,sizeof(headbuf),"X-Greeting: %s\r\n",config.greeting);
 write(cops->fd,headbuf,strlen(headbuf));

 if(need_ack) {
    write(cops->fd,telmu_headers,strlen(telmu_headers));
    /* be careful here: Some clients (Win) rely on 'X-Audio: ' *following* the
       device definitions */
    if(AUDIO_ENABLED) {
       snprintf(headbuf,sizeof(headbuf),"X-Audio-Port: %d\r\n",config.tcp_audio_port);
       write(cops->fd,headbuf,strlen(headbuf));
       }
    if(!stream) {
       /* this marks end of non-frame-specific headers for "smart" clients,
          which would otherwise expect the boundary string */
       snprintf(headbuf,sizeof(headbuf),"X-Snapshot: 1\r\n");
       write(cops->fd,headbuf,strlen(headbuf));
       }
    }

 sensor_hdr[0]=ctl_hdr[0]=0;

 if(gettimeofday(&tv1,NULL)) diag(SS_HTTP|DIERR,"gettimeofday");

 if(novideo) {
    if(!is_dup) memcpy_realloc(jpeg_buf,&my_jpeg_buf,jpeg_bytes,&my_jpeg_bufsize);
    jbytes=jpeg_bytes;
    }

 do {
    if(!novideo) {
       pthread_mutex_lock(&jpeg_buf_mtx);
       pthread_cond_wait(&new_frame_cond,&jpeg_buf_mtx);
       if(read_frame!=current_frame) read_frame=current_frame;
       else is_dup=1;
       if(!is_dup)
          memcpy_realloc(jpeg_buf,&my_jpeg_buf,jpeg_bytes,&my_jpeg_bufsize);
       jbytes=jpeg_bytes;
       pthread_mutex_unlock(&jpeg_buf_mtx);
       }

    if(!is_dup || novideo) {
       if(stream) write(cops->fd,boundary,strlen(boundary));
       if(!blanking) {
          write(cops->fd,ctype,strlen(ctype));
          snprintf(headbuf,sizeof(headbuf),xlen,jbytes);
          write(cops->fd,headbuf,strlen(headbuf));
          snprintf(headbuf,sizeof(headbuf),xidx,sent_frames);
          write(cops->fd,headbuf,strlen(headbuf));
          }
       if(sensor_hdr[0]) {
          write(cops->fd,sensor_hdr,strlen(sensor_hdr));
          sensor_hdr[0]=0;
          }
       if(ctl_hdr[0]) {
          write(cops->fd,ctl_hdr,strlen(ctl_hdr));
          ctl_hdr[0]=0;
          }
       write(cops->fd,"\r\n",2);

       if(!blanking) {
          if(write(cops->fd,my_jpeg_buf,jbytes)==-1) {
             if(errno!=EPIPE) diag(SS_HTTP|ERR,"write");
             break;
             }
          }

#if DEBUG_DUMP_IMG_LOCAL
       {
       int o;
       char fn[100];
       sprintf(fn,"/tmp/lframe%04d.jpg",sent_frames);
       if(!(o=open(fn,O_CREAT|O_WRONLY,S_IRWXU)))
          diag(SS_HTTP|DIERR,"open() DUMP_LOCAL");
       write(o,my_jpeg_buf,jbytes);
       close(o);
       }
#endif
       if(all_off) stream=0;

       if(need_ack && stream) { /* i.e. "smart" client */
          if(sensors) {
             pthread_mutex_lock(&telmu_io2http_mtx);
             if(local_seqno!=telmu_seqno || !sent_frames) {
                for(i=0;i<sensors;i++) sensor_fresh[i]=sensor[i].value;
                local_seqno=telmu_seqno;
                sensor_update=1;
                }
             pthread_mutex_unlock(&telmu_io2http_mtx);
             if(sensor_update==1) {
                int n;
                c=sensor_hdr+16; /* strlen("X-Sensor-Change:") */
                *c=0;
                sensor_tmp[0]=0;
                for(i=0;i<sensors;i++) {
                   if(sensor_fresh[i]!=sensor_current[i] || !sent_frames) {
                      sensor_current[i]=sensor_fresh[i];
                      if(!sensor_tmp[0])
                         n=snprintf(sensor_tmp,15," c%d=%d",i,sensor_current[i]);
                      else
                         n=snprintf(sensor_tmp,15,"; c%d=%d",i,sensor_current[i]);
                      if(n==-1)
                         diag(SS_HTTP|ERR,"nameval buffer overflow for sensor %d!\n",i);
                      else {
                         if(c+n>=sensor_hdr+250-3) /* EOL: \r\n\0 */
                            diag(SS_HTTP|ERR,"Sensor header buffer overflow!\n");
                         else {
                            strcat(c,sensor_tmp);
                            c+=n;
                            }
                         }
                      }
                   }
                if(sensor_hdr[16]) {
                   /* 16 = strlen("X-Sensor-Change:") */
                   strncpy(sensor_hdr,"X-Sensor-Change:",16);
                   /*diag(SS_HTTP|DEBUG,"%s\n",sensor_hdr);*/
                   strcat(c,"\r\n");
                   }
                sensor_update=0;
                }
             } /* sensors */
          
          if( (i=read_line(cops->fd,rbuf,1024,config.client_ack_timeout))<0) {
             if(i==-2) diag(SS_HTTP|DEBUG,"%s - ack timeout\n",peer_addr);
             break;
             }

          if(i>2) {
             int q, dl, sn, ii, jj, hex;
             char *s=rbuf;
             struct namevalue nv;
             rbuf[i-2]=0;
#if DEBUG_DUMP_CLIENT_DATA
             diag(SS_HTTP|DEBUG,"client data: %s\n",rbuf);
#endif
             while(s) {
                nv=parse_client_response(&s);
                if(nv.name==0) break;

                /* note: for sake of simplicity we're determining commands
                   by first letter only */
                switch(*nv.name) {

                   case 'k':
                      if(*nv.value!='1' || ctl_status!=C_DEQUEUED) break;
                      q=queue_insert(self);
                      if(q==-1) dl=-1;
                      else ctl_status=C_PENDING;
                      if(q>1) {
                         dl=config.control_timeout*(q-1);
                         sprintf(ctl_hdr,"X-Ctl-Delay: %d\r\n",dl);
                         }
                      break;

                   case 'v':
                      if(nv.value[0]=='0') blanking=1;
                      else blanking=0;
                      diag(SS_HTTP|INFO,"%s - video %sblank\n",peer_addr,blanking?"":"un");
                      break;

                   case 't':
                      for(ii=jj=0;ii<127;ii++,jj++) {
                      /* 127=sizeof(decoded_text)-1 */
                         if(!nv.value[jj]) break;
                         if(nv.value[jj]=='%') {
                            hex=0; jj++;
                            if(!nv.value[jj] || !isxdigit(nv.value[jj])) break;
                            if(nv.value[jj]<='9') hex=nv.value[jj]-'0';
                            else hex=(nv.value[jj]|0x20)-'a'+10;
                            jj++; hex<<=4;
                            if(!nv.value[jj] || !isxdigit(nv.value[jj])) break;
                            if(nv.value[jj]<='9') hex+=nv.value[jj]-'0';
                            else hex+=(nv.value[jj]|0x20)-'a'+10;
                            decoded_text[ii]=hex;
                            }
                         else decoded_text[ii]=nv.value[jj];
                         }
                      decoded_text[ii]=0;
                      diag(SS_HTTP|INFO,"%s - received text: %s\n",
                         peer_addr,decoded_text);
                      break;

                   case 'c':
                      if(ctl_status!=C_CURRENT || !sensors) break;
                      if(!isdigit(nv.name[1])) break;
                      sn=atoi(nv.name+1);
                      pthread_mutex_lock(&telmu_http2io_mtx);
                      http2io_data[sn].changed=1;
                      http2io_data[sn].value=atoi(nv.value);
#if DEBUG_DUMP_CLIENT_DATA
                      diag(SS_HTTP|DEBUG,"control %d=%u\n",sn,http2io_data[sn].value);
#endif
                      pthread_mutex_unlock(&telmu_http2io_mtx);
                      break;

                   case 'a':
                      if(ctl_status!=C_CURRENT) break;
                      diag(SS_HTTP|DEBUG,"audio request: %c, status: %d\n",*nv.value,aud_status);
                      if(*nv.value=='0' && aud_status==A_CURRENT)
                         audio_terminate_threads();
                      else if(*nv.value=='1' && aud_status==A_MUTE) {
                         aud_status=A_ACCEPTING1;
                         diag(SS_HTTP|DEBUG,"accepting audio request\n");
                         }
                      break;

                   default:
                      diag(SS_HTTP|INFO,"%s - unrecognized request: %s\n",peer_addr,nv.name);
                   } /* switch command */
                /* 't' parameter must cause end of parsing, to prevent
                   malicious users from embedding some '; x=y' stuff in text */
                if(*nv.name=='t') break;
                } /* response string scan loop */
             } /* i>2 */

          if(ctl_status==C_PENDING) {
             if(pthread_equal(thread_in_control,self)) {
                ctl_expiry=time(0)+config.control_timeout;
                ctl_status=C_CURRENT;
                aud_status=A_MUTE;
                diag(SS_HTTP|DEBUG,"%s - client gains control\n",peer_addr);
                sprintf(ctl_hdr,"X-Ctl-Expiry: %d\r\n",config.control_timeout);
                }
             }
          else if(ctl_status==C_CURRENT) {
             if(time(0)>ctl_expiry && queue_depth>1) {
                ctl_status=C_DEQUEUED;
                diag(SS_HTTP|DEBUG,"%s - client loses control\n",peer_addr);
                release_control();
                sprintf(ctl_hdr,"X-Ctl-Expiry: 0\r\n");
                queue_remove(self);
                }
             else {
                if(aud_status==A_ACCEPTING1) {
                   sprintf(ctl_hdr,"X-Audio-Start: true\r\n");
                   aud_status=A_ACCEPTING2;
                   }
                }
             }
          

          } /* need_ack */

       sent_frames++;
       tbytes+=jbytes;
       }
    if(is_dup) {
       dup_frames++;
       is_dup=0;
       usleep(40000);
       }

    if(cops->timeout) {
       if(gettimeofday(&tv2,NULL)) diag(SS_HTTP|DIERR,"gettimeofday");
       if(tv2.tv_sec-tv1.tv_sec>cops->timeout) {
          diag(SS_HTTP|INFO,"%s - connection timed out\n",peer_addr);
          break;
          }
       }
    if(novideo) { blanking=1; }
    } while(stream);

 if(ctl_status==C_PENDING || ctl_status==C_CURRENT) {
    if(ctl_status==C_CURRENT) release_control();
    queue_remove(self);
    }
 
out:
 if(gettimeofday(&tv2,NULL)) diag(SS_HTTP|DIERR,"gettimeofday");
 close(cops->fd);
 dtime=tv2.tv_sec-tv1.tv_sec+(tv2.tv_usec-tv1.tv_usec)/1000000.0;

 diag(SS_HTTP|INFO,"%s - connection terminated\n",peer_addr);

 /* avoid division by zero caused by wraparound */
 diag(SS_HTTP|DEBUG,"%s - wrote %d [+%d] frames (%llu bytes) in %.2f s - %.2f f/s, %.2f kB/s, %.2f kB/f\n",
    peer_addr,
    sent_frames,dup_frames,tbytes,dtime,
    (sent_frames)/dtime,
    (tbytes/dtime)/1024,
    sent_frames?((float)tbytes/sent_frames)/1024:0 );
 free(cops);
 free(sensor_fresh);
 free(sensor_current);
 free(my_jpeg_buf);
 decrement_clients();
 return NULL;
}


static void *capture_thread(void)
{
 float dtime;
 struct timeval tv1,tv2;

 current_frame=0;
 /* video_alloc_buffers() is now invoked in main, at startup only */
 diag(SS_MAIN|DEBUG,"Capture thread started\n");
 if(gettimeofday(&tv1,NULL)) diag(SS_MAIN|DIERR,"gettimeofday");
 while(clients) {
    video_capture_compress(nojpeg);
    //usleep(80000); // frame rate limit
#if X11MON
    if(config.viewer)
       x11_monitor(cparms.yuv2rgbbuf?cparms.yuv2rgbbuf:cparms.buffer,
          cparms.capx,cparms.capy);
#endif
    /* 'usleep' here is useless if period < 40ms (=1s/25frames) */
    /* usleep(40000); */

#if DEBUG_DUMP_IMG_GLOBAL
    {
    int o;
    char fn[100];
    sprintf(fn,"/tmp/gframe%d.jpg",current_frame);
    if(!(o=open(fn,O_CREAT|O_WRONLY,S_IRWXU)))
       diag(SS_MAIN|DIERR,"open() DUMP_GLOBAL");
    write(o,jpeg_buf,jpeg_bytes);
    close(o);
    }
#endif


    }
 if(gettimeofday(&tv2,NULL)) diag(SS_MAIN|DIERR,"gettimeofday");
 dtime=tv2.tv_sec-tv1.tv_sec+(tv2.tv_usec-tv1.tv_usec)/1000000.0;
 diag(SS_MAIN|DEBUG,
    "Capture thread stopped - %d frames in %.2f s (%.2f f/s)\n",
    current_frame,dtime,current_frame/dtime);
 pthread_cond_signal(&kill_capture_cond);
 return NULL;
}


static void usage(char *s)
{
 char *bname;
 char *u=
   "This is "VERSION"\n"
   "Usage: %s [options]\n"
   "where options are (default value in braces):\n"
   "\n  video:\n"
   "    -d DEV          capture device {"DEFAULT_VIDEO_DEVICE"}\n"
   "    -c CH           capture channel {0}\n"
   "    -s NxM          capture size {%dx%d}\n"
   "    -r FPS          capture frame rate (for some USB cameras) {%d}\n"
   "    -k X,Y,TXT      display caption TXT at position X,Y (multiple times ok)\n"
   "    -j QUAL         JPEG compression quality level {50}\n"
   "    -g              output grayscale frames\n"
   "    -f[FILE]        disable video channel; display static frame from FILE\n"
   "                    {"SYS_DIR"/"VIDEO_UNAVAIL_JPG"}\n"
   "\n  audio:\n"
   "    -i              capture device {none}\n"
   "    -o              playback device {none}\n"
   "    -a              capture and playback device {none}\n"
   "    -m              disable audio channel (both input and output)\n"
   "    -x              perform audio mixer setup\n"
   "\n  misc:\n"
   "    -C FILE         read config from FILE\n"
   "                    {"CONFIG_FILE"}\n"
   "    -q              don't dump debug info\n"
   "    -e              force logging to standard error\n"
   "    -A FAMILY       address family to listen on (any, inet, inet6)\n"
   "    -p PORT[,PORT]  TCP base (and optionally audio) port {%d,%d}\n"
   "    -S DEV          serial port {no serial link}\n"
   "    -t SEC          timeout (in seconds) for control {%d}\n"
   "    -T SEC          timeout (in seconds) for client connections {none}\n"
   "    -F[FILE]        disable all channels; display static frame from FILE\n"
   "                    {"SYS_DIR"/"VIDEO_UNAVAIL_JPG"}\n"
#ifdef X11MON
   "    -V              enable the capture monitor\n"
#endif
   "    -P SEC          execute a capture performance test SEC seconds long\n";
 bname=strrchr(s,'/');
 diag(INFO|NOTIME,
    u,(bname?bname+1:s),DEFAULT_X,DEFAULT_Y,DEFAULT_PWC_FRAME_RATE,BASE_PORT,BASE_PORT+1,DEFAULT_CONTROL_TIMEOUT);
 exit(0);
}


static void launch_thread(int stream, int sock, int timeout, struct sockaddr_storage *name, void *(*t)(void *), int clients)
{
 pthread_t thread;
 pthread_attr_t thr_attr;
 struct conn_parms *cops;
 sigset_t sigs;
 int launch_result;

 if( (cops=malloc(sizeof(struct conn_parms)))==0)
    diag(SS_MAIN|DIE,"malloc() for conn_parms failed.\n");
 cops->stream=stream; cops->fd=sock;
 cops->timeout=timeout;
 cops->sa=*name;
 cops->clients=clients;

 diag(SS_MAIN|DEBUG,"%d client(s) connected\n",clients);
 pthread_attr_init(&thr_attr);
 pthread_attr_setdetachstate(&thr_attr,PTHREAD_CREATE_DETACHED);
 sigemptyset(&sigs);
 sigaddset(&sigs,SIGHUP);
 pthread_sigmask(SIG_BLOCK,&sigs,0);
 launch_result=pthread_create(&thread,&thr_attr,t,(void*)cops);
 pthread_sigmask(SIG_UNBLOCK,&sigs,0);
 if(launch_result) {
    diag(SS_MAIN|ERR,"unable to create new server thread");
    decrement_clients();
    close(sock);
    }
}


static void handler(int signum)
{
 //diag(SS_MAIN|DEBUG,"now in handler\n");
 if(signum==SIGHUP && config.log_file) diag_restart(config.log_file);
 if(signum==SIGINT) exit(0);
}


static void set_signals(void)
{
 struct sigaction sa;

 sigemptyset(&sa.sa_mask);

 /* note that EINTR in select() must be handled explicitly anyway */
 sa.sa_flags=SA_RESTART;

 sa.sa_handler=SIG_IGN;
 sigaction(SIGPIPE,&sa,0);
 sa.sa_handler=handler;
 sigaction(SIGHUP,&sa,0);
 sa.sa_handler=handler;
 sigaction(SIGINT,&sa,0);
}


int main(int argc, char *argv[])
{
 int c;
 char *s=0, *ap, *offline_file=0, *config_file=0;
 size_t size;
 struct sockaddr_storage name;
 char peer_addr[NI_MAXHOST];
 int i,m,ss,sock;
 pthread_t capture_tid, telmu_tid;
 pthread_attr_t thr_attr;
 fd_set active_fds,read_fds;
 int fd[STREAMS];
 int setup_mixer=0;
 int performance_test=0;
 int force_log_stderr=0;
 sigset_t sigs;

 config.video_device=0;
 config.audio_in_device=config.audio_out_device=0;
 config.norm=-1;
 config.noaudio=-1;
 config.viewer=0;
 config.serial_port=0;
 config.tcp_base_port=-1;
 config.tcp_audio_port=-1;
 config.control_timeout=-1;
 config.client_timeout=-1;
 config.client_ack_timeout=-1;
 config.max_clients=-1;
 config.jpeg_quality=-1;
 config.tuner_freq=-1;
 config.greeting=0;
 config.named_pipe=0;
 config.log_file=0;
 debug_level=DEBUG;
 config.capx=-1;
 config.capy=-1;
 config.channel=-1;
 config.frame_rate=-1;
 config.video_audio_volume=-1;
 config.video_audio_input=-1;
 config.monochrome=-1;
 config.image_filters=0;
 config.address_family=-1; // all AF_* (including AF_UNSPEC) are >=0
 opterr=0;

 diag_setfile(0);
 while( (c=getopt(argc,argv,"egmqxVa:A:c:C:d:i:j:k:o:p:P:r:s:S:t:T:f::F::"))>0) switch(c) {

    case 'a':
       if(optarg==0) usage(argv[0]);
       if(!(config.audio_in_device=config.audio_out_device=strdup(optarg)))
          diag(SS_MAIN|DIE,"Out of memory\n");
       break;

    case 'A':
       if(optarg==0) usage(argv[0]);
       if(!(strcmp("any",optarg))) config.address_family=AF_UNSPEC;
       else if(!(strcmp("inet",optarg))) config.address_family=AF_INET;
       else if(!(strcmp("inet6",optarg))) config.address_family=AF_INET6;
       else diag(SS_MAIN|DIE,"Invalid address family\n");
       break;

    case 'c':
       if(optarg==0) usage(argv[0]);
       if((config.channel=atoi(optarg))<0) usage(argv[0]);
       break;

    case 'C':
       if(optarg==0) usage(argv[0]);
       config_file=strdup(optarg);
       break;

    case 'd':
       if(optarg==0) usage(argv[0]);
       if(!(config.video_device=strdup(optarg)))
          diag(SS_MAIN|DIE,"Out of memory\n");
       break;

    case 'e':
       force_log_stderr=1;
       break;

    case 'f':
    case 'F':
       offline_file=strdup(optarg?optarg:(SYS_DIR"/"VIDEO_UNAVAIL_JPG));
       novideo=1;
       if(c=='F') all_off=config.noaudio=1;
       break;

    case 'g':
       config.monochrome=1;
       break;

    case 'i':
       if(optarg==0) usage(argv[0]);
       if(!(config.audio_in_device=strdup(optarg)))
          diag(SS_MAIN|DIE,"Out of memory\n");
       break;

    case 'j':
       if(optarg==0) usage(argv[0]);
       config.jpeg_quality=atoi(optarg);
       if(config.jpeg_quality<=0 || config.jpeg_quality>100)
          diag(SS_MAIN|DIE,"JPEG quality level out of range\n");
       break;

    case 'k':
       if(optarg==0) usage(argv[0]);
       if(captions==0 && !(caption=malloc(MAX_CAPTIONS*sizeof(struct caption_t))))
          diag(SS_MAIN|DIE,"Out of memory\n");
       if(captions==MAX_CAPTIONS) {
          diag(SS_MAIN|ERROR,"Too many captions, ignoring '%s'\n",optarg);
          break;
          }
       if(!(s=strchr(optarg,','))) usage(argv[0]);
       caption[captions].x=atoi(optarg);
       caption[captions].y=atoi(s+1);
       if(!(s=strchr(s+1,','))) usage(argv[0]);
       if(!(caption[captions].raw=strdup(s+1)))
          diag(SS_MAIN|DIE,"Out of memory\n");
       if(!(caption[captions].cooked=malloc(CAPTION_SIZEOF_COOKED)))
          diag(SS_MAIN|DIE,"Out of memory\n");
       captions++;
       break;

    case 'm':
       config.noaudio=1;
       break;

    case 'o':
       if(optarg==0) usage(argv[0]);
       if(!(config.audio_out_device=strdup(optarg)))
          diag(SS_MAIN|DIE,"Out of memory\n");
       break;

    case 'p':
       if(optarg==0) usage(argv[0]);
       config.tcp_base_port=atoi(optarg);
       if(config.tcp_base_port<=0 || config.tcp_base_port>65535)
          diag(SS_MAIN|DIE,"Invalid base port `%d'\n",config.tcp_base_port);
       if( (ap=strchr(optarg,',')) ) {
          config.tcp_audio_port=atoi(ap+1);
          if(config.tcp_audio_port<=0 || config.tcp_audio_port>65535)
             diag(SS_MAIN|DIE,"Invalid audio port `%d'\n",config.tcp_audio_port);
          }
       break;

    case 'P':
       if(optarg==0) usage(argv[0]);
       performance_test=atoi(optarg);
       if(performance_test<=0)
          diag(SS_MAIN|DIE,"Invalid performance test duration\n");
       break;

    case 'q':
       debug_level=DEBUG+1;
       break;

    case 'r':
       if(optarg==0) usage(argv[0]);
       config.frame_rate=atoi(optarg);
       break;

    case 's':
       if(optarg==0 || (s=strchr(optarg,'x'))==0) usage(argv[0]);
       if( (config.capx=atoi(optarg))<=0 || (config.capy=atoi(s+1))<=0 )
          usage(argv[0]);
       break;

    case 'S':
       if(optarg==0) usage(argv[0]);
       if(!(config.serial_port=strdup(optarg)))
          diag(SS_MAIN|DIE,"Out of memory\n");
       break;

    case 't':
       if(optarg==0) usage(argv[0]);
       if((config.control_timeout=atoi(optarg))<0)
          diag(SS_MAIN|DIE,"Invalid control timeout\n");
       break;

    case 'T':
       if(optarg==0) usage(argv[0]);
       if((config.client_timeout=atoi(optarg))<0)
          diag(SS_MAIN|DIE,"Invalid client timeout\n");
       break;

    case 'x':
       setup_mixer=1;
       break;

    case 'V':
#ifdef X11MON
       config.viewer=1;
#else
       diag(SS_MAIN|DIE,"Compiled without support for capture monitor\n");
#endif
       break;

    default:
       usage(argv[0]);
       break;
    }

#if 0
 /* should we need non-option args... */
 argc-=optind;
 argv+=optind;
 if(argc>1) ...=atoi(argv[1]);
#endif

 if(parse_config_file(config_file?config_file:CONFIG_FILE,config_file?1:0))
    diag(SS_MAIN|DIE,"malformed config file\n");

 if(UNSET_STRING(config.video_device)) config.video_device=DEFAULT_VIDEO_DEVICE;
 if(UNSET_INT(config.channel)) config.channel=0;
 if(UNSET_INT(config.frame_rate)) config.frame_rate=DEFAULT_PWC_FRAME_RATE;
 if(UNSET_INT(config.video_audio_input)) config.video_audio_input=0;
 if(UNSET_INT(config.video_audio_volume)) config.video_audio_volume=65535;
 if(UNSET_INT(config.norm)) config.norm=VIDEO_MODE_PAL;
 if(UNSET_INT(config.noaudio)) config.noaudio=0;
 if(UNSET_INT(config.monochrome)) config.monochrome=0;
 if(UNSET_INT(config.address_family)) config.address_family=AF_UNSPEC;
 if(UNSET_INT(config.tcp_base_port)) config.tcp_base_port=BASE_PORT;
 if(UNSET_INT(config.tcp_audio_port))
    config.tcp_audio_port=config.tcp_base_port+1;
 if(config.tcp_audio_port>65535)
    diag(SS_MAIN|DIE,"Cannot set audio port to base port + 1\n");
 if(config.tcp_audio_port==config.tcp_base_port)
    diag(SS_MAIN|DIE,"Base port and audio port must be different\n");
 if(UNSET_INT(config.max_clients)) config.max_clients=5;
 if(UNSET_INT(config.control_timeout))
    config.control_timeout=DEFAULT_CONTROL_TIMEOUT;
 if(UNSET_INT(config.client_timeout)) config.client_timeout=0;
 if(UNSET_INT(config.client_ack_timeout)) config.client_ack_timeout=30;
 if(UNSET_INT(config.jpeg_quality)) config.jpeg_quality=50;
 if(UNSET_INT(config.capx)) { config.capx=DEFAULT_X; config.capy=DEFAULT_Y; }
 if(UNSET_STRING(config.greeting)) config.greeting="Experimental Palantir server";
 if(force_log_stderr) config.log_file=0;
 if(performance_test) {
    debug_level=DEBUG;
    if(novideo)
       diag(SS_MAIN|DIE,"Performance test requested with video capture disabled!\n");
    }

 set_signals();

 if(config.log_file) diag_setfile(config.log_file);
 diag(SS_MAIN|INFO,"-- "VERSION" starting --\n");
 diag(SS_MAIN|INFO,"Definitions for %d devices found\n",sensors);

 if(!config.serial_port) diag(SS_MAIN|DEBUG,"No serial port specified\n");
 else diag(SS_MAIN|DEBUG,"Using serial port at %s\n",config.serial_port);

 if(!config.named_pipe) diag(SS_MAIN|DEBUG,"No named pipe specified\n");
 else diag(SS_MAIN|DEBUG,"Using named pipe at %s\n",config.named_pipe);

 jpeg_init();
 if(!novideo) {
    video_init(config.video_device,config.channel,config.monochrome);
    video_alloc_buffers();
#ifdef X11MON
    if(config.viewer) x11_monitor_init(cparms.capx,cparms.capy);
#endif
    }
 else jpeg_load_file(offline_file);

 /* we're going to use thr_attr a lot, but we need to init it only once */
 pthread_attr_init(&thr_attr);
 pthread_attr_setdetachstate(&thr_attr,PTHREAD_CREATE_DETACHED);

 if(performance_test) {
    diag(SS_MAIN|INFO,"*** Running a %d-second performance test\n",performance_test);
    clients=1;
    if(pthread_create(&capture_tid,&thr_attr,(void*(*)(void*))capture_thread,(void*)0))
       diag(SS_MAIN|DIE,"Unable to create capture thread\n");
    sleep(performance_test);
    clients=0;
    sleep(2);
    exit(0);
    }
 if(config.viewer) {
    diag(SS_MAIN|INFO,"*** Running as viewer only\n");
    nojpeg=1;
    clients=1;
    capture_thread();
    }

 queue_init();
 telmu_init(config.serial_port,config.named_pipe);

 /* the telmu thread should only be started when actually useful. */
 if(sensors) {
    diag(SS_MAIN|DEBUG,"Starting device thread\n");
    sigemptyset(&sigs);
    sigaddset(&sigs,SIGHUP);
    pthread_sigmask(SIG_BLOCK,&sigs,0);
    if(pthread_create(&telmu_tid,&thr_attr,(void*(*)(void*))telmu_io,(void*)0))
       diag(SS_MAIN|DIERR,"unable to create device thread");
    pthread_sigmask(SIG_UNBLOCK,&sigs,0);
    }

 if(AUDIO_ENABLED) {
    if(audio_init(config.audio_in_device,config.audio_out_device,setup_mixer)) {
       diag(SS_MAIN|ERROR,"Audio initialization failed. Audio channel will be disabled\n");
       config.noaudio=1;
       }
    }

 for(i=0;i<STREAMS;i++) {
    if(i==AUDIO_STREAM && !AUDIO_ENABLED) continue;
    fd[i]=create_sock(i==VIDEO_STREAM?config.tcp_base_port:config.tcp_audio_port,config.address_family);
    }
 size=sizeof(name);
 FD_ZERO(&active_fds);
 for(i=0,m=0;i<STREAMS;i++) {
    if(i==AUDIO_STREAM && !AUDIO_ENABLED) continue;
    if(listen(fd[i],(i==VIDEO_STREAM?MAX_VIDEO_CONNECTIONS:1))<0)
       diag(SS_MAIN|DIERR,"listen()");
    if(fd[i]>m) m=fd[i];
    FD_SET(fd[i],&active_fds);
    }

 diag(SS_MAIN|DEBUG,"Control queue timeout set to %d s\n",config.control_timeout);
 for(;;) {
    /* diag(SS_MAIN|INFO,"sleeping on select()\n"); */
    read_fds=active_fds;
    do {
       ss=select(1+m,&read_fds,0,0,NULL);
       if(ss<0 && errno!=EINTR) diag(SS_MAIN|DIERR,"select() failed");
       } while(ss<0);
    for(i=0;i<STREAMS;i++) {
       if(i==AUDIO_STREAM && !AUDIO_ENABLED) continue;
       if(FD_ISSET(fd[i],&read_fds)) {
          sock=accept(fd[i],(struct sockaddr *)&name,&size);
          if(sock<0) {
             diag(SS_MAIN|ERR,"accept() for stream #%d (port %d) failed",
                i,i==VIDEO_STREAM?config.tcp_base_port:config.tcp_audio_port);
             continue;
             }

          if(getnameinfo((struct sockaddr *)&name,sizeof(name),peer_addr,sizeof(peer_addr),0,0,NI_NUMERICHOST|NI_NUMERICSERV)!=0) {
             /*strcpy(sbuf,"unknown");*/
             strcpy(peer_addr,"unknown");
             }
          diag(SS_MAIN|INFO,"%s - #%d: connection accepted, spawning server.\n",peer_addr,i);
          if(i==VIDEO_STREAM) {
             int cls, launch_result=0;
             pthread_mutex_lock(&clients_mtx);
             cls=++clients;
             /*diag(SS_MAIN|DEBUG,"%d client(s) connected\n",clients);*/
             if(!novideo && clients==1) {
                 sigemptyset(&sigs);
                 sigaddset(&sigs,SIGHUP);
                 pthread_sigmask(SIG_BLOCK,&sigs,0);
                 if((launch_result=pthread_create(&capture_tid,&thr_attr,(void*(*)(void*))capture_thread,(void*)0))!=0) clients--;
                 pthread_sigmask(SIG_UNBLOCK,&sigs,0);
                 }
             pthread_mutex_unlock(&clients_mtx);

             if(launch_result) {
                diag(SS_MAIN|ERR,"Unable to create capture thread");
                close(sock);
                }
             else launch_thread(i,sock,config.client_timeout,&name,(void*(*)(void*))http_server,cls);
             }
          else if(i==AUDIO_STREAM) {
             if(!AUDIO_ENABLED) {
                close(sock);
                diag(SS_MAIN|INFO,"Audio channel disabled!\n");
                }
             else {
                if(!DEBUG_ALLOW_ALL_AUDIO_REQS && aud_status!=A_ACCEPTING2) {
                   close(sock);
                   diag(SS_MAIN|INFO,"Unexpected connection to audio channel\n");
                   }
                else audio_launch_threads(sock,&name);
                }
             }
          }
       } /* for streams */
    }

 /* as to now, this cleanup sequence gets never invoked */
 if(!novideo) video_cleanup();
 telmu_cleanup();
 /* audio cleanup should take place only if audio_init was invoked.
    (anyway it does no harm.)
    Maybe use an extra var to keep track of audio init status?  (must
    propagate it to http_server) */
 audio_cleanup();
 jpeg_cleanup();
}
