
/*
    Control queue management functions.

    Copyright 2000-2007 by David Santinoli <david@santinoli.com>
*/

#include <pthread.h>
#include <malloc.h>
#include "queue.h"
#include "diag.h"

static const char rcsid[]="$Id: queue.c,v 1.6 2007/05/18 23:23:46 santinod Exp $";

static struct queue_item *queue_head;

pthread_mutex_t queue_mtx=PTHREAD_MUTEX_INITIALIZER;
pthread_t thread_in_control;
int queue_depth;

#define DEBUG_DUMP_QUEUE 0

void queue_init(void)
{
 queue_depth=0;
 queue_head=0;
}


struct queue_item *queue_create_item(pthread_t tid)
{
 struct queue_item *tmp;
 if( (tmp=malloc(sizeof(struct queue_item))) ) {
    tmp->next=0;
    tmp->tid=tid;
    }
 return tmp;
}


#if DEBUG_DUMP_QUEUE
void queue_print(void)
{
 struct queue_item *ii;
 int i=0;

 ii=queue_head;
 while(ii) {
    diag(SS_MAIN|DEBUG,"Queue [%d]: %d\n",i++,ii->tid);
    ii=ii->next;
    }
}
#endif


void queue_remove(pthread_t tid)
{
 struct queue_item *prev, *ii;
 int found=0;

 pthread_mutex_lock(&queue_mtx);
 ii=prev=queue_head;
 while(ii && !pthread_equal(tid,ii->tid)) {
    prev=ii;
    ii=ii->next;
    }
 if(pthread_equal(tid,ii->tid)) {
    found=1;
    if(ii==queue_head) {
       queue_head=queue_head->next;
       free(ii);
       /* note: we used to set thread_in_control to a special value (0) if
	  queue_head==0 (empty queue), to make the "Am I in control?" test
          in http threads fail for sure.
	  However, that test is only executed in the C_PENDING state,
          which implies that the queue is non-empty - so there's no
          point in that assignment.
          (Besides, it's not elegant to make assumption about allowed and
          available values for an opaque type.)
          */
       if(queue_head) thread_in_control=queue_head->tid;
       }
    else {
       prev->next=ii->next;
       free(ii);
       }
    queue_depth--;
    }
 diag(SS_MAIN|DEBUG,"control queue depth %d after removal\n",queue_depth);
#if DEBUG_DUMP_QUEUE
 queue_print();
#endif

 pthread_mutex_unlock(&queue_mtx);

 if(!found) diag(SS_MAIN|ERROR,"inconsistent control queue (dequeue)");
} 


/*
   return no of items in control queue, including the newly inserted one, or
   -1 upon trouble.
*/
int queue_insert(pthread_t tid)
{
 struct queue_item *ii, *tmp;
 int qd;

 tmp=queue_create_item(tid);
 if(!tmp) {
    diag(SS_MAIN|ERR,"Could not create control queue item");
    return -1;
    }

 pthread_mutex_lock(&queue_mtx);

 if(!queue_head) {
    queue_head=tmp;
    thread_in_control=tid;
    }
 else {
    ii=queue_head;
    while(ii->next) ii=ii->next;
    ii->next=tmp;
    }

 qd=++queue_depth;
#if DEBUG_DUMP_QUEUE
 queue_print();
#endif

 pthread_mutex_unlock(&queue_mtx);

 diag(SS_MAIN|DEBUG,"control queue depth %d after insertion\n",queue_depth);
 return qd;
}
