
/*
   Telemetry subsystem (a tribute to Apollo 13)

   Copyright 2000-2007 by David Santinoli <david@santinoli.com>
*/

#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include "main.h"
#include "diag.h"
#include "video.h"
#include "parser.h"
#include "sensor_def.h"
#include "internal.h"

#define SERIAL_SPEED B9600
#define FDS 2
#define DEBUG_DUMP_IO 0

static const char rcsid[]="$Id: telmu.c,v 1.14 2007/05/18 23:24:36 santinod Exp $";

static int serial_fd=-1;
static int pipe_fd=-1;
static struct termios saved_tm;
extern int novideo;

struct http2io_data_def *http2io_data=0;
static struct http2io_data_def *local_io_data;

int carrier_remap[3][CARRIER_SUBCHANS];

int telmu_seqno=1; /* whereas in http servers is set to 0 */
pthread_mutex_t telmu_io2http_mtx=PTHREAD_MUTEX_INITIALIZER;

pthread_mutex_t telmu_http2io_mtx=PTHREAD_MUTEX_INITIALIZER;
char *telmu_headers="";

static char *encode(char *s)
{
 static char buf[300];
 char *p=buf;

 while(*s) {
    if((*s==';' || *s==' ' || *s=='%') && p+3<buf+300) {
       *p++='%';
       *p++=(*s/16<10 ? '0'+*s/16 : 'A'+*s/16);
       *p++=((*s&15)<10 ? '0'+(*s&15) : 'A'+(*s&15));
       s++;
       }
    else *p++=*s++;
    }
 *p=0;
 return buf;
}


void telmu_init(char *serial_port, char *named_pipe)
{
 struct termios tm;
 int i, flags;
 char tmp[300];

 if(serial_port && config.need_serial) {
    if((serial_fd=open(serial_port,O_RDWR|O_NONBLOCK))==-1)
       diag(SS_SENSOR|DIERR,"Cannot open serial port %s",serial_port);
   
    if(!isatty(serial_fd))
       diag(SS_SENSOR|DIE,"%s doesn't appear to be a serial line.\n");
   
    /* set non-blocking mode for I/O */
    flags=fcntl(serial_fd,F_GETFL);
    if(flags==-1) diag(SS_SENSOR|DIERR,"unable to get serial line flags");
    if(fcntl(serial_fd,F_SETFL,flags|O_NONBLOCK)==-1)
       diag(SS_SENSOR|DIERR,"unable to set serial line flags");
   
    if(tcgetattr(serial_fd,&saved_tm)==-1) diag(SS_SENSOR|DIERR,"tcgetattr()");
    if(tcgetattr(serial_fd,&tm)==-1) diag(SS_SENSOR|DIERR,"tcgetattr()");
    cfmakeraw(&tm);
    if(cfsetspeed(&tm,SERIAL_SPEED)==-1) diag(SS_SENSOR|DIERR,"cfsetspeed()");
    if(tcsetattr(serial_fd,TCSANOW,&tm)==-1) diag(SS_SENSOR|DIERR,"tcsetattr()");
    diag(SS_SENSOR|INFO,"Serial port %s initialized\n",serial_port);
    }

 if(named_pipe && config.need_pipe) {
    if((pipe_fd=open(named_pipe,O_RDWR))==-1)
       diag(SS_SENSOR|DIERR,"Cannot open named pipe `%s'",named_pipe);
    /* should we make sure that the pipe is drained, or is it just nonsense? */

    flags=fcntl(pipe_fd,F_GETFL);
    if(flags==-1) diag(SS_SENSOR|DIERR,"unable to get FIFO file flags");
    if(fcntl(pipe_fd,F_SETFL,flags|O_NONBLOCK)==-1)
       diag(SS_SENSOR|DIERR,"unable to set FIFO file flags");
    }

 /* build appropriate "X-Sensor" headers for sensor definitions.
    Return a statically allocated string to be used by http threads. */

 telmu_headers=malloc(3000);
 sprintf(telmu_headers,"X-Defined-Sensors: %d\r\n",sensors);
 for(i=0;i<sensors;i++) {
    strcat(telmu_headers,"X-Sensor-Define: ");
    if(*sensor[i].name) {
       sprintf(tmp,"name=%s; ",encode(sensor[i].name));
       strcat(telmu_headers,tmp);
       }
    if(*sensor[i].hint) {
       sprintf(tmp,"hint=%s; ",encode(sensor[i].hint));
       strcat(telmu_headers,tmp);
       }
    sprintf(tmp,"imin=%d; imax=%d; fmin=%s; fmax=%s; dir=%d; type=%c\r\n",
       sensor[i].imin,sensor[i].imax,
       sensor[i].cfmin,sensor[i].cfmax,
       sensor[i].dir,sensor[i].widget);
    strcat(telmu_headers,tmp);
    }

 if(sensors) {
    http2io_data=(struct http2io_data_def *)malloc(sensors*sizeof(struct http2io_data_def));
    if(http2io_data) http2io_data[0].changed=0;

    local_io_data=(struct http2io_data_def *)malloc(sensors*sizeof(struct http2io_data_def));
    }
}


void *telmu_io(void *none)
{
 fd_set read_fds, active_fds;
 int i, n, s, dev, max_fd, fd[2];
 int update_seqno;
 unsigned char buf[3];
 struct timeval timeout;
 struct {
    int buffed;
    unsigned char buf[3];
    } read_buf[FDS];

 /* check if any of fd's failed to open */
 fd[LINK_PIPE]=pipe_fd;
 fd[LINK_SERIAL]=serial_fd;

 /* might be worth unrolling... */
 FD_ZERO(&active_fds);
 for(i=0,max_fd=0;i<FDS;i++) {
    if(fd[i]>max_fd) max_fd=fd[i];
    if(fd[i]>=0) FD_SET(fd[i],&active_fds);
    read_buf[i].buffed=0;
    }

 while(1) {

    /* .changed is set to 1 for HTTP data change request, to 2 if there's
       feedback from device (and to 3 if there are both; feedback from
       device determines value in this case).
       Keeping track of the modification source prevents us from feeding
       devices with data we've just read from them. */

    timeout.tv_sec=0;
    timeout.tv_usec=300000;
    read_fds=active_fds;
    do {
       s=select(1+max_fd,&read_fds,0,0,&timeout);
       if(s<0 && errno!=EINTR) diag(SS_SENSOR|DIERR,"select() failed");
       } while(s<0);

    pthread_mutex_lock(&telmu_http2io_mtx);
    for(i=0; i<sensors; i++) {
       local_io_data[i]=http2io_data[i];
#if DEBUG_DUMP_IO
       if(http2io_data[i].changed) diag(SS_SENSOR|DEBUG,"http: c%d=%d\n",i,local_io_data[i].value);
#endif
       http2io_data[i].changed=0;
       }
    pthread_mutex_unlock(&telmu_http2io_mtx);

    if(s>0) { /* there are data in the queues */
       for(i=0;i<FDS;i++) {
          if(fd[i]>=0 && FD_ISSET(fd[i],&read_fds)) {
             if((n=read(fd[i],read_buf[i].buf+read_buf[i].buffed,
                3-read_buf[i].buffed))<=0) break;
#if DEBUG_DUMP_IO
             diag(SS_SENSOR|DEBUG,"read %d bytes from fd %d\n",n,i);
#endif

             if((read_buf[i].buffed+=n)==3) {

#if DEBUG_DUMP_IO
                diag(SS_SENSOR|DEBUG,"packet from fd %d: %02x %02x %02x\n",i,
                   read_buf[i].buf[0],read_buf[i].buf[1],read_buf[i].buf[2]);
#endif
                dev=carrier_remap[i][read_buf[i].buf[0]];
                if(dev>=0 && dev<sensors) {
                   local_io_data[dev].changed|=2;
                   local_io_data[dev].value=
                      (read_buf[i].buf[1]<<8)+read_buf[i].buf[2];
                   }
                read_buf[i].buffed=0;
                }
             }
          }

       } /* read data */


    for(i=0;i<sensors;i++) {
       /* never trust clients! Upon receiving a HTTP update request,
	  only sync after checking that the device is writable and the
          value is in range */
       if((local_io_data[i].changed&3)==1 && 
             local_io_data[i].value>=sensor[i].imin &&
             local_io_data[i].value<=sensor[i].imax &&
             sensor[i].dir!=SD_R) {
          if(sensor[i].link==LINK_INTERNAL) {
          /* trap local controls */
             if(sensor[i].channel==10)
                midi_prog_change(local_io_data[i].value);
             else if(!novideo)
                video_change_param(sensor[i].channel,
                   local_io_data[i].value);
             continue;
             }

          else if(sensor[i].link==LINK_SERIAL || sensor[i].link==LINK_PIPE) {
             buf[0]=sensor[i].channel;
             buf[1]=local_io_data[i].value>>8;
             buf[2]=local_io_data[i].value&0xff;
             if(sensor[i].link==LINK_SERIAL && serial_fd>=0) {
#if DEBUG_DUMP_IO
                diag(SS_SENSOR|DEBUG,"packet to fd %d: %02x %02x %02x\n",serial_fd,buf[0],buf[1],buf[2]);
#endif
                if((n=write(serial_fd,buf,3))<3)
                   diag(SS_SENSOR|ERR,"incomplete write to serial device (%d bytes)\n",n);
                }
             else if(sensor[i].link==LINK_PIPE && pipe_fd>=0) {
#if DEBUG_DUMP_IO
                diag(SS_SENSOR|DEBUG,"packet to fd %d: %02x %02x %02x\n",pipe_fd,buf[0],buf[1],buf[2]);
#endif
                if((n=write(pipe_fd,buf,3))<3)
                   diag(SS_SENSOR|ERR,"incomplete write to fifo (%d bytes)\n",n);
                }
             }
          } /* if changed */
       } /* for */

    /* Update state
      (should do this only if any changes are actually there?) */

       update_seqno=0;
       pthread_mutex_lock(&telmu_io2http_mtx);
       for(i=0;i<sensors;i++)
          if( (  ((local_io_data[i].changed&1)&&sensor[i].notifier==NT_CLIENT)
              || ((local_io_data[i].changed&2)&&sensor[i].notifier==NT_DEVICE))
               && sensor[i].value!=local_io_data[i].value) {
             sensor[i].value=local_io_data[i].value;
#if DEBUG_DUMP_IO
             diag(SS_SENSOR|DEBUG,"updating c%d=%d\n",i,local_io_data[i].value);
#endif
             update_seqno=1;
             }
       if(update_seqno) telmu_seqno++;
       pthread_mutex_unlock(&telmu_io2http_mtx);


    } /* while */
 return (void*)0;
}


void telmu_cleanup(void)
{
 if(serial_fd>=0) {
    /* restore pristine settings */
    if(tcsetattr(serial_fd,TCSANOW,&saved_tm)==-1) diag(SS_SENSOR|DIERR,"tcsetattr()");
    close(serial_fd);
    }
 if(pipe_fd>=0) close(pipe_fd);
}
