
/*
   Feed the Palantir local FIFO with some system hard/soft stats.

   This file is part of the Palantir suite.
   Copyright 2000-2005 by David Santinoli

   Usage: sysfeed [ named_pipe [ base_address ] ]

*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/sysinfo.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <string.h>

#ifndef NAMED_PIPE
#define NAMED_PIPE "/usr/local/share/palantir/telmu_pipe"
#endif

/* This is where lm_sensors data are read from. Customize the paths according
   to your hardware. */
#define I2C_CPUFAN_DEVICE "/sys/bus/i2c/devices/2-0290/fan1_input"
#define I2C_CPUTEMP_DEVICE "/sys/bus/i2c/devices/2-0290/temp2_input"

static const char rcsid[] = "$Id: sysfeed.c,v 1.2 2005/08/30 22:22:26 santinod Exp $";

char i2cbuf[128];

char *read_i2c_device(int fd)
{
 if(lseek(fd,0,SEEK_SET)==-1) {
    fprintf(stderr,
       "Warning: unable to lseek i2c device `"I2C_CPUFAN_DEVICE"'\n");
    return 0;
    }

 if(read(fd,i2cbuf,127)==-1) {
    fprintf(stderr,
       "Warning: unable to read from i2c device `"I2C_CPUFAN_DEVICE"'\n");
    return 0;
    }
 //fprintf(stderr,"read `%s'\n",c);
 return i2cbuf;
}


int main(int argc, char *argv[])
{
 char buf[100], *named_pipe=NAMED_PIPE;
 unsigned int v;
 int base=0, t=0;
 int ff, retrying=0;
 int cpufan_fd, cputemp_fd;
 struct sysinfo si;
 struct stat st;

 if(argc==2 && (!strcmp(argv[1],"--help") || !strcmp(argv[1],"-h")) ) {
    fprintf(stderr,"Usage: %s [ named_pipe [ base_address ] ]\n",argv[0]);
    exit(0);
    }
 signal(SIGPIPE,SIG_IGN);
 if(argc>1) named_pipe=argv[1];
 if(argc>2) base=atoi(argv[2]);
 if(lstat(named_pipe,&st)==-1 || !S_ISFIFO(st.st_mode)) {
    fprintf(stderr,"Unable to locate FIFO file at %s\n",named_pipe);
    exit(1);
    }
 if( (ff=open(named_pipe,O_WRONLY))<0 ) {
    fprintf(stderr,"Cannot open FIFO file %s\n",named_pipe);
    exit(1);
    }
 if( (cpufan_fd=open(I2C_CPUFAN_DEVICE,O_RDONLY) )<0)
    fprintf(stderr,
       "Warning: unable to open i2c device `"I2C_CPUFAN_DEVICE"'\n");
 if( (cputemp_fd=open(I2C_CPUTEMP_DEVICE,O_RDONLY) )<0)
    fprintf(stderr,
       "Warning: unable to open i2c device `"I2C_CPUTEMP_DEVICE"'\n");

 errno=0;
 for(;;) {
    if((t%5)<3 && sysinfo(&si)==-1) break;
    switch(t%5) {
       case 0:
          v=si.loads[0];
          if(v>65535) v=65535;
          //printf("load=%d\n",v);
          break;

       case 1:
          v=si.procs;
          break;

       case 2:
          v=(1-((float)si.freeram)/si.totalram)*65535;
          break;

       case 3:
          if(cpufan_fd<0 || !read_i2c_device(cpufan_fd)) {
             t++;
             continue;
             }
          if(sscanf(i2cbuf,"%d",&v)!=1) {
             fprintf(stderr,
                "Warning: garbage from i2c device `"I2C_CPUFAN_DEVICE"'\n");
             t++;
             continue;
             }
          break;

       case 4:
          if(cputemp_fd<0 || !read_i2c_device(cputemp_fd)) {
             t++;
             continue;
             }
          if(sscanf(i2cbuf,"%d",&v)!=1) {
             fprintf(stderr,
                "Warning: garbage from i2c device `"I2C_CPUTEMP_DEVICE"'\n");
             t++;
             continue;
             }
          v/=1000;
          break;
       }

    buf[0]=base+(t%5);
    buf[1]=(v>>8)&0xff;
    buf[2]=v&0xff;
    if(write(ff,buf,3)==-1) {
       if(errno==EPIPE) {
          if(retrying==0) {
             fprintf(stderr,"Reading end apparently disconnected. Will keep retrying every 10 seconds...\n");
             retrying=1;
             }
          sleep(10);
          continue;
          }
       break;
       }

    if(retrying) {
       fprintf(stderr,"Reading end reconnected.\n");
       retrying=0;
       }
    sleep(2);
    t++;
    }
 return 0;
}
