
/*
    Interface to v4l.

    Copyright 2000-2007 by David Santinoli <david@santinoli.com>
*/

#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#include <linux/videodev.h>
#include <unistd.h>
#include <errno.h>
#include <sys/mman.h>
#include "config.h"
#include "diag.h"
#include "jpeg.h"
#include "video.h"
#include "yuv2rgb.h"
#include "parser.h"
#include "frequencies.h"
#include "pwc-ioctl.h"
#include "pwc-mpt.h"
#include "caption.h"
#include <linux/meye.h>

static const char rcsid[]="$Id: video.c,v 1.26 2007/05/18 23:25:52 santinod Exp $";
static int philips_cam=0, meye_cam=0, has_tuners=0, channels;
static int has_ptz=0; /* 1: recent drivers, 2: legacy */
static int prev_capx=0, prev_capy=0;

static int palette_yuv=0;
/* 0: use VIDEO_PALETTE_RGB24; 1: use VIDEO_PALETTE_YUV* */

struct capture_parms cparms;

#define X_MONO_PALETTE  VIDEO_PALETTE_GREY
#define X_MONO_PALETTE_DEPTH  8


void meye_setup(void)
{
 struct meye_params par;

 if(ioctl(cparms.fd,MEYEIOC_G_PARAMS,&par)<0) {
    diag(SS_VIDEO|ERR,"MEYEIOC_G_PARAMS");
    return;
    }
 par.subsample=(cparms.capx==640&&cparms.capy==480)?0:1;
 par.quality=7;
 par.sharpness=32;
 par.agc=48;
 par.picture=0;
 par.framerate=0;
 if(ioctl(cparms.fd,MEYEIOC_S_PARAMS,&par)<0)
    diag(SS_VIDEO|ERR,"MEYEIOC_S_PARAMS");
 diag(SS_VIDEO|DEBUG,"meye-specific setup terminated.\n");
}


void video_init(char *devname, int chan, int mono)
{
 int i;
 char *caps[]={"capture","tuner","teletext","overlay","chromakey","clipping","frameram","scales","monochrome","subcapture","mpeg_decoder","mpeg_encoder","mjpeg_decoder","mjpeg_encoder"};
 struct video_capability v_cap;
 struct video_channel v_chan;
 struct video_tuner v_tuner;
 struct video_audio v_audio;
 struct video_picture v_pic;
 unsigned long freq;

 if((cparms.fd=open(devname,O_RDWR))==-1)
    diag(SS_VIDEO|DIERR,"Cannot open video device %s",devname);

 if(ioctl(cparms.fd,VIDIOCGCAP,&v_cap))
    diag(SS_VIDEO|ERR,"GCAP");
 else {
    diag(SS_VIDEO|DEBUG,"Card:\t%s (%s)\n",v_cap.name,devname);

    /* cam detection */
    if(!strcmp(v_cap.name,"meye")) meye_cam=1;
    else if(sscanf(v_cap.name,"Philips %d webcam",&i)>=1) philips_cam=1;
    else {
       struct pwc_probe probe;
       if(ioctl(cparms.fd,VIDIOCPWCPROBE,&probe)==0 &&
             !strcmp(v_cap.name,probe.name))
          philips_cam=1;
       }
    }

 diag(SS_VIDEO|DEBUG,"Capabilities:\t%d (",v_cap.type);
 for(i=0;i<32;i++)
    if(v_cap.type&(1<<i))
       diag(DEBUG|NOTIME,"%s ",(i<sizeof(caps)?caps[i]:"unknown_cap") );
 diag(DEBUG|NOTIME,")\n");

 diag(SS_VIDEO|DEBUG,"Size:\t(%dx%d)-(%dx%d)\n",v_cap.minwidth,v_cap.minheight,v_cap.maxwidth,v_cap.maxheight);

 channels=v_cap.channels;
 /* When queried with ioctl(VIDIOCGCHAN) with a non-existing
    video_channel.channel, some drivers rewrite it instead of
    letting ioctl fail. We have to detect this situation. */
 for(i=0;i<channels;i++) {
    v_chan.channel=i;
    if(ioctl(cparms.fd,VIDIOCGCHAN,&v_chan)) diag(SS_VIDEO|DIERR,"GCHAN");
    diag(SS_VIDEO|DEBUG,"Channel no. %d ('%s') ",v_chan.channel,v_chan.name);
    diag(DEBUG|NOTIME,"\ttuners: %d, flags: %d, type: %d, norm: %d\n",
       v_chan.tuners,v_chan.flags,v_chan.type,v_chan.norm);
    if(v_chan.tuners) has_tuners=1;
    }

 if(chan>channels)
    diag(SS_VIDEO|DIE,"Channel must be in range [0..%d]\n",channels);
 else {
    v_chan.channel=chan;
    if(ioctl(cparms.fd,VIDIOCSCHAN,&v_chan)) diag(SS_VIDEO|DIERR,"SCHAN");
    }

 if(has_tuners) {
    v_tuner.tuner=0;
    ioctl(cparms.fd,VIDIOCGTUNER,&v_tuner);
    v_tuner.mode=config.norm;
    ioctl(cparms.fd,VIDIOCSTUNER,&v_tuner);

    if(!UNSET_FLOAT(config.tuner_freq)) {
       freq=config.tuner_freq*16;
       diag(SS_VIDEO|DEBUG,"Setting tuner freq to %.3f\n",config.tuner_freq);
       if(ioctl(cparms.fd,VIDIOCSFREQ,&freq))
          diag(SS_VIDEO|DIERR,"set tuner freq");
       }
    }

 if(v_cap.audios && !config.noaudio) {
    if(config.video_audio_input>=v_cap.audios)
       diag(SS_VIDEO|DIERR,"Audio input %d does not exist on selected video device",config.video_audio_input);

    for(i=0;i<v_cap.audios;i++) {
       v_audio.audio=config.video_audio_input;
       if(ioctl(cparms.fd,VIDIOCGAUDIO,&v_audio))
          diag(SS_VIDEO|DIERR,"get video device audio");
       if(i!=config.video_audio_input) {
          if(v_audio.flags&VIDEO_AUDIO_MUTABLE) {
             v_audio.flags|=VIDEO_AUDIO_MUTE;
             if(ioctl(cparms.fd,VIDIOCSAUDIO,&v_audio))
                diag(SS_VIDEO|DIERR,"Failed to mute audio input %d",i);
   	  diag(SS_VIDEO|DEBUG,"Muted audio input %d\n",i);
   	  }
          else {
             diag(SS_VIDEO|DEBUG,"Unable to mute audio input %d\n",i);
             if(v_audio.flags&VIDEO_AUDIO_VOLUME) {
   	     v_audio.volume=0;
                if(ioctl(cparms.fd,VIDIOCSAUDIO,&v_audio))
                   diag(SS_VIDEO|DIERR,"Failed to set volume for audio input %d",i);
   	     diag(SS_VIDEO|DEBUG,"Volume for audio input %d set to 0\n",i);
   	     }
             else diag(SS_VIDEO|DEBUG,"Audio input %d does not support volume control\n",i);
   	  }
          }
       else {
          v_audio.flags&=~VIDEO_AUDIO_MUTE;
          if(ioctl(cparms.fd,VIDIOCSAUDIO,&v_audio))
             diag(SS_VIDEO|DIERR,"Failed to unmute audio input %d",i);
          diag(SS_VIDEO|DEBUG,"Unmuted audio input %d\n",i);
   
          if(v_audio.flags&VIDEO_AUDIO_VOLUME) {
             v_audio.volume=config.video_audio_volume;
             if(ioctl(cparms.fd,VIDIOCSAUDIO,&v_audio))
                diag(SS_VIDEO|DIERR,"Failed to set volume for audio input %d",i);
             diag(SS_VIDEO|DEBUG,"Volume for audio input %d set to %d\n",i,config.video_audio_volume);
             }
          else diag(SS_VIDEO|DEBUG,"Audio input %d does not support volume control\n",i);
          }
       }
    }

 /* image props setup */
 if(ioctl(cparms.fd,VIDIOCGPICT,&v_pic)) diag(SS_VIDEO|DIERR,"GPICT");
 v_pic.palette=(mono?X_MONO_PALETTE:VIDEO_PALETTE_RGB24);
 v_pic.depth=cparms.depth=(mono?X_MONO_PALETTE_DEPTH:24);
 cparms.mono=mono;
 v_pic.brightness=32768;
 v_pic.contrast=32768;
 v_pic.colour=32768;
 v_pic.hue=32768;
 v_pic.whiteness=32768;
 
 if(ioctl(cparms.fd,VIDIOCSPICT,&v_pic)) {
    if(!mono) {
       palette_yuv=1;
       diag(SS_VIDEO|DEBUG,"RGB palette unavailable.\n");

       diag(SS_VIDEO|DEBUG,"trying YUV420P...\n");
       v_pic.palette=VIDEO_PALETTE_YUV420P;
       v_pic.depth=cparms.depth=12;
       if(ioctl(cparms.fd,VIDIOCSPICT,&v_pic)==0) goto pal_out;

       diag(SS_VIDEO|DEBUG,"trying YUV422...\n");
       v_pic.palette=VIDEO_PALETTE_YUV422;
       v_pic.depth=cparms.depth=16;
       if(ioctl(cparms.fd,VIDIOCSPICT,&v_pic))
          diag(SS_VIDEO|DIERR,"SPICT");

       pal_out: ;
       }
    else diag(SS_VIDEO|DIERR,"SPICT (mono palette might be unavailable)");
    }
 if(ioctl(cparms.fd,VIDIOCGPICT,&v_pic)) diag(SS_VIDEO|DIERR,"GPICT");
 diag(SS_VIDEO|DEBUG,"brightness: %d  hue: %d  colour: %d\n",v_pic.brightness,v_pic.hue,v_pic.colour);
 diag(SS_VIDEO|DEBUG,"contrast: %d  whiteness: %d\n",v_pic.contrast,v_pic.whiteness);
 /* Note: CGPICT can return a depth of 24 even with YUV palette - odd. */
 diag(SS_VIDEO|DEBUG,"depth: (%d)  palette: %d\n",v_pic.depth,v_pic.palette);

 cparms.palette=v_pic.palette;
 cparms.capx=config.capx;
 cparms.capy=config.capy;

 if(meye_cam) meye_setup();
 // x11_monitor_init(&cparms);
}


/*
   mmap() (or allocate the buffer for read()ing from) the video device.
   Allocate the yuv2rgb buffer if needed.
   Previously mmap()ed/malloc()ed memory is freed before reallocation upon
   change of requested capture size (runtime change of capture size is not
   currently implemented. Changes should take place before restarting capture
   thread).
*/
void video_alloc_buffers(void)
{
 struct video_window v_win;

 if(cparms.capx==prev_capx && cparms.capy==prev_capy) return;

 if(prev_capx && prev_capy) {
    video_free_buffers();
    jpeg_free_rows();
    }

 /* capture window setup */
 if(!meye_cam) {
    if(ioctl(cparms.fd,VIDIOCGWIN,&v_win)) diag(SS_VIDEO|DIERR,"GWIN");
    }
 v_win.x=v_win.y=0;
 v_win.width=cparms.capx;
 v_win.height=cparms.capy;
 prev_capx=cparms.capx; prev_capy=cparms.capy;
 if(philips_cam && (v_win.flags & PWC_FPS_MASK)) {
    v_win.flags&=~PWC_FPS_FRMASK;
    v_win.flags|=(config.frame_rate<<PWC_FPS_SHIFT);
    }
 else v_win.flags=0;
 v_win.chromakey=0;
 v_win.clips=0;
 v_win.clipcount=0;
 if(!meye_cam) {
    if(ioctl(cparms.fd,VIDIOCSWIN,&v_win)) {
       if(errno==EINVAL)
          diag(SS_VIDEO|DIE,"Invalid capture size and/or frame rate.\n");
       else diag(SS_VIDEO|DIERR,"SWIN");
       }
    if(ioctl(cparms.fd,VIDIOCGWIN,&v_win)) diag(SS_VIDEO|DIERR,"GWIN");
    /* diag(SS_VIDEO|DEBUG,"Requested capture box is %dx%d\n",config.capx,config.capy); */
    diag(SS_VIDEO|DEBUG,"Actual capture box is %dx%d (requested %dx%d)\n",v_win.width,v_win.height,config.capx,config.capy);
    }
 if(philips_cam) {
    struct l_pwc_mpt_range pwc_range;
    if(!ioctl(cparms.fd,VIDIOCPWCMPTGRANGE,&pwc_range)) has_ptz=1;
    else if(errno==EINVAL) {
       diag(SS_VIDEO|INFO,"Pan/tilt unsupported by the driver in use. Retrying in legacy mode...\n");
       if(!ioctl(cparms.fd,L_VIDIOCPWCMPTGRANGE,&pwc_range)) has_ptz=2;
       }
    if(!has_ptz) {
       switch(errno) {
          case ENXIO:
             diag(SS_VIDEO|INFO,"Camera has no pan/tilt control\n");
             break;
          case EINVAL:
             diag(SS_VIDEO|INFO,"Pan/tilt unsupported by the driver in use.\n");
             break;
          default:
             diag(SS_VIDEO|ERR,"VIDIOCPWCMPTGRANGE");
          }
       }
    else {
       unsigned int ptz_reset=3;
       diag(SS_VIDEO|DEBUG,"Pan/tilt ranges: %d...%d, %d...%d\n",
       pwc_range.pan_min,pwc_range.pan_max,
       pwc_range.tilt_min,pwc_range.tilt_max);
       if(1) {
          diag(SS_VIDEO|DEBUG,"Resetting PTZ devices\n");
          ioctl(cparms.fd,VIDIOCPWCMPTRESET,&ptz_reset);
          /* sleep 1-2s? */
          }
       }

    diag(SS_VIDEO|DEBUG,"Frame rate set to %d f/s\n",(v_win.flags&PWC_FPS_MASK)>>PWC_FPS_SHIFT);
    }

 /* test mmap() availability */
 cparms.mbuffer=0;
 if(!ioctl(cparms.fd,VIDIOCGMBUF,&cparms.v_mbuf)) {
    diag(SS_VIDEO|DEBUG,"Video buffer size: %d, frames: %d\n",
       cparms.v_mbuf.size,cparms.v_mbuf.frames);
#if 0
    { int i;
    for(i=0;i<cparms.v_mbuf.frames;i++)
       diag(SS_VIDEO|DEBUG,"\tFrame %d offset %d\n",i,cparms.v_mbuf.offsets[i]);
    }
#endif
    /* 2.4:
    cparms.mbuffer=mmap(0,cparms.v_mbuf.size,PROT_READ,MAP_SHARED,cparms.fd,0);
    */
    /* 2.6: (works for 2.4 too) */
    cparms.mbuffer=mmap(0,cparms.v_mbuf.size,PROT_READ|PROT_WRITE,MAP_SHARED,cparms.fd,0);
    if((unsigned char *)-1==cparms.mbuffer) {
       diag(SS_VIDEO|ERR,"mmap of %d bytes failed",cparms.v_mbuf.size);
       cparms.mbuffer=0;
       }
    }
 if(cparms.mbuffer==0)
    diag(SS_VIDEO|DEBUG,"mmap() not available - reverting to read()\n");

 cparms.bufsize=(cparms.depth*cparms.capx*cparms.capy)/8;
 if((cparms.buffer=(unsigned char *)malloc(cparms.bufsize))==0)
    diag(SS_VIDEO|DIE,"malloc() of %d bytes failed",cparms.bufsize);

 if(palette_yuv) {
    diag(SS_VIDEO|DEBUG,"Allocating yuv->rgb buffer (%d bytes)\n",3*(cparms.capx)*(cparms.capy));
    cparms.yuv2rgbbuf=malloc(3*(cparms.capx)*(cparms.capy));
    /* if YUV->RGB needs a look-up table, initialize it here. */
    }
 else cparms.yuv2rgbbuf=0;

 if(prev_capx && prev_capy)
    jpeg_compress_init(cparms.capx,cparms.capy,
       (cparms.yuv2rgbbuf?cparms.yuv2rgbbuf:cparms.buffer),
       cparms.mono,config.jpeg_quality);
}


void video_change_param(int par, int val)
{
 struct video_picture v_pic;
 struct video_channel v_chan;
 int fr;

 if(par>=0 && par<3) {
    if(ioctl(cparms.fd,VIDIOCGPICT,&v_pic)) diag(SS_VIDEO|DIERR,"GPICT");
    switch(par) {
       case 0:
          v_pic.brightness=val;
          /* diag(SS_VIDEO|DEBUG,"brightness set to %d\n",val); */
          break;
       case 1:
          v_pic.contrast=val;
          /* diag(SS_VIDEO|DEBUG,"contrast set to %d\n",val); */
          break;
       case 2:
          v_pic.colour=val;
          /* diag(SS_VIDEO|DEBUG,"colour set to %d\n",val); */
          break;
          }
    if(ioctl(cparms.fd,VIDIOCSPICT,&v_pic)) diag(SS_VIDEO|DIERR,"SPICT");
    }
 else if(par==3 && has_tuners) {
    fr=(freqs[val%(sizeof(freqs)/sizeof(float))]*16);
    if(ioctl(cparms.fd,VIDIOCSFREQ,&fr))
       diag(SS_VIDEO|DIERR,"set tuner freq");
    }
 else if(par==4 && val>=0 && val<=channels) {
    /* must do VIDIOCGCHAN first, so we have a valid v_chan.norm */
    if(ioctl(cparms.fd,VIDIOCGCHAN,&v_chan)) diag(SS_VIDEO|ERR,"GCHAN");
    v_chan.channel=val;
    if(ioctl(cparms.fd,VIDIOCSCHAN,&v_chan)) diag(SS_VIDEO|ERR,"SCHAN");
    }
 else if((par==5 || par==6) && has_ptz) {
    struct l_pwc_mpt_angles pwc_angles;
    int req=0; // just to avoid "uninit'd var warning"

    if(has_ptz==1) req=VIDIOCPWCMPTGANGLE;
    else if(has_ptz==2) req=L_VIDIOCPWCMPTGANGLE;

    if(ioctl(cparms.fd,req,&pwc_angles))
       diag(SS_VIDEO|ERR,"PWCMPTGANGLE");
    pwc_angles.absolute=1;
    if(par==5) pwc_angles.pan=val;
    else if(par==6) pwc_angles.tilt=val;
    //diag(SS_VIDEO|INFO,"PTZ: %d=%d\n",par,val);

    if(has_ptz==1) req=VIDIOCPWCMPTSANGLE;
    else if(has_ptz==2) req=L_VIDIOCPWCMPTSANGLE;
    if(ioctl(cparms.fd,req,&pwc_angles))
       diag(SS_VIDEO|ERR,"PWCMPTSANGLE");
    }
}


void video_capture_compress(int nojpeg)
{
 struct video_mmap v_mmap;
 static int first_frame=1, frameno=0;
 int e;

 v_mmap.frame=frameno;
 v_mmap.width=cparms.capx;
 v_mmap.height=cparms.capy;
 v_mmap.format=cparms.palette;

 /* NOTE! With some cards, read() may block indefinitely. Use mmap() where
  * possible. */
 if(!cparms.mbuffer) {
    if(read(cparms.fd,cparms.buffer,cparms.bufsize)==-1) {
       diag(SS_VIDEO|ERR,"read() from video device failed");
       if(errno==ENODEV) diag(SS_VIDEO|DIE,"Device disconnected\n");
       }
    }
 else {
    if(first_frame) {
       if(ioctl(cparms.fd,VIDIOCMCAPTURE,&v_mmap)) {
          diag(SS_VIDEO|ERR,"CMCAPTURE - mmap failed, should revert to read()");
          if(errno==ENODEV) diag(SS_VIDEO|DIE,"Device disconnected\n");
          }
       first_frame=0;
       }
    v_mmap.frame=(frameno+1)%cparms.v_mbuf.frames;
    if(ioctl(cparms.fd,VIDIOCMCAPTURE,&v_mmap)) {
       diag(SS_VIDEO|ERR,"CMCAPTURE - mmap failed, should revert to read()");
       if(errno==ENODEV) diag(SS_VIDEO|DIE,"Device disconnected\n");
       }

    do {
       e=ioctl(cparms.fd,VIDIOCSYNC,&frameno);
       } while(e && errno==EINTR);
    if(e) {
       diag(SS_VIDEO|ERR,"CSYNC");
       if(errno==ENODEV) diag(SS_VIDEO|DIE,"Device disconnected\n");
       }
    memcpy(cparms.buffer,cparms.mbuffer+cparms.v_mbuf.offsets[frameno],
       cparms.bufsize);
    frameno=v_mmap.frame;
    }

#if 0
/*
  BGR24: 0x42475218 (IMGFMT_BGR24)
  YUV:   0x30323449 (IMGFMT_I420)
  mplayer -rawvideo on:w=320:h=240:size=230400:format=0x42475218 named_pipe
*/
 {
 static int f=0;
 if(!f) f=open("out.raw",O_WRONLY);
 //f=open("out.raw",O_CREAT|O_WRONLY|O_TRUNC,0644);
 write(f,cparms.buffer,3*(cparms.capx)*(cparms.capy));
 //close(f);
 }
#endif

 if(palette_yuv) yuv2rgb(cparms.palette,cparms.yuv2rgbbuf,cparms.buffer,cparms.capx,cparms.capy);
 else {
    int x,y;
    unsigned char c, *p=cparms.buffer, *p1, *p2;
    for(y=0;y<cparms.capy;y++)
       if(config.image_filters&IMAGE_FILTER_FLIP)
          for(x=0,p1=(p+=3*cparms.capx),p2=p1+3*cparms.capx-1;x<cparms.capx/2;x++) {
             c=*(p1); *(p1++)=*(p2); *(p2--)=c;
             c=*(p1); *p1++=*(p2); *p2--=c;
             c=*(p1); *p1++=*p2; *(p2--)=c;
             }
       else
          for(x=0;x<cparms.capx;x++,p+=3) {
             c=*(p+2);
             *(p+2)=*(p);
             *(p)=c;
             }
    }

 draw_captions(cparms.yuv2rgbbuf?cparms.yuv2rgbbuf:cparms.buffer,v_mmap.width,v_mmap.height);

 if(!nojpeg) jpeg_compress(cparms.capx,cparms.capy);

}


void video_free_buffers(void)
{
 free(cparms.buffer);
 if(cparms.mbuffer)
    munmap(cparms.mbuffer,cparms.v_mbuf.size);

 if(palette_yuv) free(cparms.yuv2rgbbuf);
}


void video_cleanup(void)
{
 video_free_buffers();
 close(cparms.fd);
}
